# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

from monai.transforms.utils import get_transform_backends, print_transform_backends


class TestPrintTransformBackends(unittest.TestCase):

    def test_get_number_of_conversions(self):
        tr_t_or_np, *_ = get_transform_backends()
        self.assertGreater(len(tr_t_or_np), 0)
        print_transform_backends()


if __name__ == "__main__":
    a = TestPrintTransformBackends()
    a.test_get_number_of_conversions()
